#include <vector>
#include <iostream>

#include <isa/yaml.h>
#include <isa/pdg.h>
#include "da.h"

extern "C" {
#include "isl_util.h"
}
#include "isl/constraint.h"
#include <isl/set.h>
#include <isl/flow.h>

using pdg::PDG;
using namespace std;

namespace da {

/* A pair of a node and an access from that node.
 * "map" is the converted access relation.
 */
struct na_pair {
    pdg::node 	*node;
    pdg::access	*access;
    isl_map	*map;
    na_pair(pdg::node *n, pdg::access *a) : node(n), access(a), map(NULL) {}
    ~na_pair() {
	isl_map_free(map);
    }
};

static int precedes_level_nodes(na_pair *first, na_pair *second);
static int precedes_level_accesses(na_pair *first, na_pair *second);

/* Given a map from a domain to an orthogonal projection of an array
 * (say, the rows of an array), mapping i to m(i), this function
 * extends the range of the mapping to the original array and extends
 * the domain of the mapping correspondingly such that (i,j) maps
 * to (m(i),j), with (m(i),j) identifying an element of the array.
 * The bounds on j are taken from the size of the array.
 *
 * The mapping from i to (i,j) is stored in the "extension" field
 * of the access.
 *
 * The dependences computed using these extended access mappings,
 * will map a (possibly) extended source domain to a (possibly)
 * extended sink domain.  One or both of these domains need to
 * be transformed back to the original domains using the inverse
 * of the corresponding extensions.
 */
static isl_map *extend_access(isl_map *map, na_pair *na)
{
	pdg::array *array = na->access->array;
	assert(isl_map_n_out(map) < array->dims.size());
	unsigned s_dim = array->dims.size() - isl_map_n_out(map);
	isl_id *array_id = NULL;
	if (isl_map_has_tuple_id(map, isl_dim_out))
		array_id = isl_map_get_tuple_id(map, isl_dim_out);
	isl_space *dim = isl_map_get_space(map);
	dim = isl_space_drop_dims(dim, isl_dim_in,
				0, isl_space_dim(dim, isl_dim_in));
	dim = isl_space_drop_dims(dim, isl_dim_out,
				0, isl_space_dim(dim, isl_dim_out));
	dim = isl_space_add_dims(dim, isl_dim_in, s_dim);
	dim = isl_space_add_dims(dim, isl_dim_out, s_dim);
	isl_basic_map *id = isl_basic_map_identity(isl_space_copy(dim));
	isl_local_space *ls = isl_local_space_from_space(dim);
	isl_int v;
	isl_int_init(v);
	for (int i = 0; i < s_dim; ++i) {
		isl_constraint *c;
		c = isl_inequality_alloc(isl_local_space_copy(ls));
		isl_int_set_si(v, 1);
		isl_constraint_set_coefficient(c, isl_dim_out, 0, v);
		id = isl_basic_map_add_constraint(id, c);
		c = isl_inequality_alloc(isl_local_space_copy(ls));
		isl_int_set_si(v, -1);
		isl_constraint_set_coefficient(c, isl_dim_out, 0, v);
		isl_int_set_si(v, array->dims[isl_map_n_out(map)+i]-1);
		isl_constraint_set_constant(c, v);
		id = isl_basic_map_add_constraint(id, c);
	}
	isl_local_space_free(ls);
	isl_int_clear(v);
	map = isl_map_product(map, isl_map_from_basic_map(id));
	map = isl_map_flatten_range(map);
	if (array_id)
		map = isl_map_set_tuple_id(map, isl_dim_out, array_id);
	if (!na->access->extension) {
		isl_map *ext = isl_map_copy(map);
		ext = isl_set_unwrap(isl_map_domain(ext));
		ext = isl_map_reverse(isl_map_domain_map(ext));
		na->access->extension = new pdg::IslMap(ext);
	}
	if (!na->access->extended_map)
		na->access->extended_map = new pdg::IslMap(isl_map_copy(map));
	return map;
}

/* If access "access" contains any nested accesses, then the domain
 * of the access relation contains extra dimensions corresponding to
 * the values of the nested accesses.
 * Add these extra dimensions, with ranges given by the value_bounds
 * of the corresponding array to domain "dom".
 * If a nested access array does not have value_bounds, then we assume
 * an infinite interval.
 */
static isl_set *append_nested_value_domains(isl_set *dom, pdg::access *access)
{
	isl_ctx *ctx;

	ctx = isl_set_get_ctx(dom);
	for (int i = 0; i < access->nested.size(); ++i) {
		pdg::call_or_access *coa = access->nested[i];
		assert(coa->type == pdg::call_or_access::t_access);
		pdg::access *nested = coa->access;
		isl_set *bounds;
		if (nested->array->value_bounds)
			bounds = nested->array->value_bounds->get_isl_set(ctx);
		else {
			isl_space *dim = isl_set_get_space(dom);
			dim = isl_space_drop_dims(dim, isl_dim_set,
					    0, isl_space_dim(dim, isl_dim_set));
			dim = isl_space_add_dims(dim, isl_dim_set, 1);
			bounds = isl_set_universe(dim);
		}
		dom = isl_set_product(dom, bounds);
	}
	return dom;
}

/* Combine constraints of the "pure" mapping with the constraints
 * on the domain.  If the range of the mapping is of a dimension
 * that is lower than the dimension of the accessed array,
 * we extend the dimension of both domain and range of the mapping
 * with the missing dimension.  The size of domain and range
 * in these dimensions is set to the extent of the array in the
 * corresponding missing dimension.   Each point in the original
 * domain is therefore expanded to a hyperrectangle and each point
 * in this hyperrectangle is mapped onto a single point in the array.
 *
 * If node->source is a wrapped map, then the iteration domain
 * is the domain of this map.
 */
static isl_map *convert_access(na_pair *na)
{
	isl_map *map = na->access->map->get_isl_map();
	isl_set *dom = na->node->source->get_isl_set();

	if (isl_set_is_wrapping(dom))
		dom = isl_map_domain(isl_set_unwrap(dom));

	dom = append_nested_value_domains(dom, na->access);
	if (isl_map_n_in(map) != isl_set_dim(dom, isl_dim_set))
		assert(0);
	map = isl_map_intersect_domain(map, dom);
	if (isl_map_n_out(map) != na->access->array->dims.size())
		map = extend_access(map, na);
	return map;
}

/* Add parameters to "set" identifying the last iteration of the access
 * identified by "na".
 *
 * In particular, we add a parameter
 *
 *	__last_<stmt>_<access_nr>_valid >= 1
 *
 * and parameters
 *
 *	__last_<stmt>_<access_nr>_<i> = it_<i>
 *
 * with i ranging over the iterators, the dimensions of the set.
 * "na" is attached to the first parameter, so that it can be recovered
 * in refine_controls().  If the set already references some of these
 * parameters, then we don't add the parameter again, but instead
 * simply add the corresponding constraint.
 *
 * We add the constraint *_valid >= 1 instead of *_valid = 1,
 * because we don't want the negation of this constraint to result
 * in two parts (*_valid <= 0 and *_valid >= 2).
 */
static __isl_give isl_set *add_parametrization(__isl_take isl_set *set,
	na_pair *na)
{
	int pos;
	int n_source;
	isl_ctx *ctx;
	const char *stmt_name;
	isl_id *id;
	char name[60];

	ctx = isl_set_get_ctx(set);
	n_source = isl_set_dim(set, isl_dim_out);
	stmt_name = isl_set_get_tuple_name(set);
	snprintf(name, sizeof(name), "__last_%s_%d_valid",
					stmt_name, na->access->nr);
	pos = isl_set_find_dim_by_name(set, isl_dim_param, name);
	if (pos < 0) {
		id = isl_id_alloc(ctx, name, na);
		pos = isl_set_dim(set, isl_dim_param);
		set = isl_set_add_dims(set, isl_dim_param, 1);
		set = isl_set_set_dim_id(set, isl_dim_param, pos, id);
	}
	set = isl_set_lower_bound_si(set, isl_dim_param, pos, 1);

	for (int i = 0; i < n_source; ++i) {
		snprintf(name, sizeof(name), "__last_%s_%d_%d",
			 stmt_name, na->access->nr, i);

		pos = isl_set_find_dim_by_name(set, isl_dim_param, name);
		if (pos < 0) {
			id = isl_id_alloc(ctx, name, NULL);
			pos = isl_set_dim(set, isl_dim_param);
			set = isl_set_add_dims(set, isl_dim_param, 1);
			set = isl_set_set_dim_id(set, isl_dim_param, pos, id);
		}

		set = isl_set_equate(set, isl_dim_param, pos, isl_dim_set, i);
	}

	return set;
}

/* Is the i-th parameter of "map" a control, i.e., a parameter
 * introduced by add_parametrization()?
 * In particular, is the parameter of the form __last_*?
 */
static bool is_control(__isl_keep isl_map *map, int i)
{
	const char *name;
	const char *prefix = "__last_";
	size_t prefix_len = strlen(prefix);

	if (!isl_map_has_dim_id(map, isl_dim_param, i))
		return false;
	name = isl_map_get_dim_name(map, isl_dim_param, i);
	return strncmp(name, prefix, prefix_len) == 0;
}

/* Remove all controls that are redundant, i.e., that do not appear
 * in any of the constraints.
 * Set *has_controls to true if there are any controls that are not redundant.
 */
static __isl_give isl_map *remove_redundant_controls(__isl_take isl_map *dep,
	bool *has_controls)
{
	int i;
	int n_param;

	*has_controls = false;

	n_param = isl_map_dim(dep, isl_dim_param);
	for (i = n_param - 1; i >= 0; --i) {
		if (!is_control(dep, i))
			continue;
		if (isl_map_involves_dims(dep, isl_dim_param, i, 1))
			*has_controls = true;
		else
			dep = isl_map_project_out(dep, isl_dim_param, i, 1);
	}

	return dep;
}

/* Rename controls of "dep" from
 *
 *	__last_<source_stmt>_<source_acc_nr>_*
 *
 * to
 *
 *	__last_<source_stmt>_<source_acc_nr>_<sink_stmt>_<sink_acc_nr>_*
 *
 * "access_nr" is the sink access nr.
 */
static __isl_give isl_map *rename_controls(__isl_take isl_map *dep,
	int access_nr)
{
	char buf[100];
	int n_param;
	const char *name;
	const char *underscore;

	n_param = isl_map_dim(dep, isl_dim_param);
	for (int i = 0; i < n_param; ++i) {
		int len;
		if (!is_control(dep, i))
			continue;
		name = isl_map_get_dim_name(dep, isl_dim_param, i);
		underscore = strrchr(name, '_');
		assert(underscore);
		len = underscore + 1 - name;
		memcpy(buf, name, len);
		snprintf(buf + len, sizeof(buf) - len, "%s_%d_%s",
			isl_map_get_tuple_name(dep, isl_dim_out),
			access_nr, name + len);
		dep = isl_map_set_dim_name(dep, isl_dim_param, i, buf);
	}

	return dep;
}

extern "C" {
	static int extract_dep(__isl_take isl_map *dep, int must,
		void *dep_user, void *user);
}

/* Extract the single dependence relation from the result of
 * dataflow analyis and assign it to *user.
 */
static int extract_dep(__isl_take isl_map *dep, int must, void *dep_user,
	void *user)
{
	isl_map **dep_p = (isl_map **) user;
	assert(!*dep_p);
	*dep_p = dep;
	return 0;
}

/* Add parameters to "map" identifying the last iteration of the source
 * if node->source involves any controls.
 * "na" identifies the access corresponding to the source.
 */
static __isl_give isl_map *add_parametrization(__isl_take isl_map *map,
	na_pair *na)
{
	isl_set *set;

	if (!isl_set_is_wrapping(na->node->source->set))
		return map;

	set = isl_set_universe(isl_space_domain(isl_map_get_space(map)));
	set = add_parametrization(set, na);
	map = isl_map_intersect_domain(map, set);

	return map;
}

/* Produce a set with constraints on the parameters identifying
 * the last iteration of write_na that can be used to eliminate
 * some cases.
 *
 * "id" is the valid bit of the last iteration
 * "write_na" represents the source
 * "read_na" represents the sink
 * "precedes_level" is the comparison routine that was used during
 *	the dependence analysis
 *
 * In particular, produce the set
 *
 *	(id = 1 and there is a memory dependence from write_na to read_na) or
 *	id = 0
 *
 * The result of the memory based dependence analysis will have some
 * constraints on the parameters identifying the last source iteration.
 */
static __isl_give isl_set *refinement(__isl_keep isl_id *id, na_pair *write_na,
	na_pair *read_na, isl_access_level_before precedes_level)
{
	isl_access_info *acc;
	isl_flow *deps;
	isl_map *write_map;
	isl_map *dep = NULL;
	isl_set *ref;
	isl_set *invalid;
	int pos;

	acc = isl_access_info_alloc(isl_map_copy(read_na->map),
				    read_na, precedes_level, 1);
	write_map = isl_map_copy(write_na->map);
	write_map = add_parametrization(write_map, write_na);
	acc = isl_access_info_add_source(acc, write_map, 0, write_na);
	deps = isl_access_info_compute_flow(acc);
	isl_flow_foreach(deps, &extract_dep, &dep);
	isl_flow_free(deps);

	assert(dep);

	ref = isl_map_range(dep);

	pos = isl_set_find_dim_by_id(ref, isl_dim_param, id);
	assert(pos >= 0);
	ref = isl_set_fix_si(ref, isl_dim_param, pos, 1);

	invalid = isl_set_universe(isl_set_get_space(ref));
	invalid = isl_set_fix_si(invalid, isl_dim_param, pos, 0);
	ref = isl_set_union(ref, invalid);

	return ref;
}

/* Eliminate all parameters that match the first len characters
 * of valid, except valid itself.
 */
static __isl_give isl_basic_map *eliminate_controls(
	__isl_take isl_basic_map *bmap, const char *valid, size_t len)
{
	int n_param;

	n_param = isl_basic_map_dim(bmap, isl_dim_param);
	for (int i = 0; i < n_param; ++i) {
		const char *name;
		if (!isl_basic_map_has_dim_id(bmap, isl_dim_param, i))
			continue;
		name = isl_basic_map_get_dim_name(bmap, isl_dim_param, i);
		if (!strcmp(name, valid))
			continue;
		if (strncmp(name, valid, len))
			continue;
		bmap = isl_basic_map_eliminate(bmap, isl_dim_param, i, 1);
	}

	return bmap;
}

extern "C" {
	static int basic_drop_constraints_on_invalid_controls(
		__isl_take isl_basic_map *bmap, void *user);
}

/* If some __last_*_valid is set to zero, then eliminate
 * the corresponding __last_*_* iterators, keeping only __last_*_valid.
 */
static int basic_drop_constraints_on_invalid_controls(
	__isl_take isl_basic_map *bmap, void *user)
{
	isl_map **res = (isl_map **) user;
	int n_param;
	const char *prefix = "__last_";
	size_t prefix_len = strlen(prefix);
	const char *suffix = "_valid";
	size_t suffix_len = strlen(suffix);
	isl_int v;

	isl_int_init(v);

	n_param = isl_basic_map_dim(bmap, isl_dim_param);
	for (int i = 0; i < n_param; ++i) {
		const char *name;
		size_t len;

		if (!isl_basic_map_has_dim_id(bmap, isl_dim_param, i))
			continue;
		name = isl_basic_map_get_dim_name(bmap, isl_dim_param, i);
		if (strncmp(name, prefix, prefix_len))
			continue;
		len = strlen(name);
		if (len < suffix_len)
			continue;
		if (strcmp(name + len - suffix_len, suffix))
			continue;
		if (!isl_basic_map_plain_is_fixed(bmap, isl_dim_param, i, &v))
			continue;
		if (isl_int_cmp_si(v, 0) != 0)
			continue;
		bmap = eliminate_controls(bmap, name, len - suffix_len);
	}

	isl_int_clear(v);

	*res = isl_map_union(*res, isl_map_from_basic_map(bmap));

	return 0;
}

static __isl_give isl_map *drop_constraints_on_invalid_controls(
	__isl_take isl_map *dep)
{
	isl_map *res;

	res = isl_map_empty(isl_map_get_space(dep));

	isl_map_foreach_basic_map(dep,
		&basic_drop_constraints_on_invalid_controls, &res);

	isl_map_free(dep);

	return res;
}

/* Try and simplify the constraints on the parameters identifying
 * last source iterations in "dep".
 *
 * "read_na" represents the sink
 * "precedes_level" is the comparison routine that was used during
 *	the dependence analysis
 *
 * We look for parameters called __last_* that have non-NULL user field.
 * In practice, those are the parameters __last_*_valid, created
 * in add_parametrization().
 * For each of those parameters we call refinement() to obtain
 * constraints on the corresponding tuple of parameters.
 *
 * After collecting all these constraints, we add them to "dep".
 * Finally, we drop all constraints on the last source iterators
 * in those subsets where the valid bit is off and coalesce.
 */
static __isl_give isl_map *refine_controls(__isl_take isl_map *dep,
	na_pair *read_na, isl_access_level_before precedes_level)
{
	int n_param;
	const char *prefix = "__last_";
	size_t prefix_len = strlen(prefix);
	isl_set *ref;

	ref = isl_set_universe(isl_space_range(isl_map_get_space(dep)));

	n_param = isl_map_dim(dep, isl_dim_param);
	for (int i = 0; i < n_param; ++i) {
		const char *name;
		isl_id *id;
		na_pair *write_na;
		isl_set *ref_i;

		if (!isl_map_has_dim_id(dep, isl_dim_param, i))
			continue;
		name = isl_map_get_dim_name(dep, isl_dim_param, i);
		if (strncmp(name, prefix, prefix_len))
			continue;
		id = isl_map_get_dim_id(dep, isl_dim_param, i);
		write_na = (na_pair *) isl_id_get_user(id);
		if (!write_na) {
			isl_id_free(id);
			continue;
		}

		ref_i = refinement(id, write_na, read_na, precedes_level);
		ref = isl_set_intersect(ref, ref_i);

		isl_id_free(id);
	}

	dep = isl_map_intersect_range(dep, ref);
	dep = drop_constraints_on_invalid_controls(dep);
	dep = isl_map_coalesce(dep);

	return dep;
}

/* Remove all parameters that were introduced by add_parametrization().
 */
static __isl_give isl_map *remove_all_controls(__isl_take isl_map *dep)
{
	int i;
	int n_param;
	const char *name;

	n_param = isl_map_dim(dep, isl_dim_param);
	for (i = n_param - 1; i >= 0; --i) {
		if (!is_control(dep, i))
			continue;
		dep = isl_map_project_out(dep, isl_dim_param, i, 1);
	}
	dep = isl_map_coalesce(dep);

	return dep;
}

struct add_dep_info {
	PDG				*pdg;
	pdg::array			*a;
	type				t;
	enum pdg::dependence::type	dtype;
	na_pair				*read_na_pair;
	isl_access_level_before		precedes_level;
};

/* Simplify the constraints on the parameters introduced
 * in add_parametrization().
 * We first remove all those parameters that turn out not be needed.
 * If there are any of those parameters
 * left, the we try to eliminate some impossible cases and rename
 * the parameters to also include a reference to the sink.
 * The resulting relation is assigned to controlled_relation,
 * while the relation field is assigned the result of projecting out
 * all those parameters.
 */
static __isl_give isl_map *simplify_controls(__isl_take isl_map *dep,
	add_dep_info *info, bool *has_controls)
{
	bool hs;

	if (!has_controls)
		has_controls = &hs;
	dep = remove_redundant_controls(dep, has_controls);
	if (*has_controls) {
		dep = refine_controls(dep, info->read_na_pair,
					info->precedes_level);
		dep = rename_controls(dep, info->read_na_pair->access->nr);
	}

	return dep;
}

/* Extract a single dependence from the result of dataflow analysis.
 *
 * We first simplify the constraints on the parameters introduced
 * in add_parametrization().
 */
static int add_dep(__isl_take isl_map *dep, int must, void *dep_user, void *user)
{
	bool has_controls;
	na_pair *write_na = (na_pair *) dep_user;
	add_dep_info *info = (struct add_dep_info *)user;
	isl_ctx *ctx = info->pdg->get_isl_ctx();
	pdg::dependence *d;

	dep = isl_map_coalesce(dep);
	dep = simplify_controls(dep, info, &has_controls);
	if (isl_map_is_empty(dep)) {
		isl_map_free(dep);
		return 0;
	}

	d = new pdg::dependence;
	d->array = info->a;
	d->type = info->dtype;
	d->from = write_na->node;
	d->to = info->read_na_pair->node;
	d->from_access = write_na->access;
	d->to_access = info->read_na_pair->access;

	if (d->from_access->extension || d->to_access->extension)
		d->extended_relation =
			new pdg::IslMap(remove_all_controls(isl_map_copy(dep)));
	if (d->from_access->extension)
		dep = isl_map_apply_domain(dep,
			    isl_map_reverse(
				d->from_access->extension->get_isl_map(ctx)));
	if (d->to_access->extension)
		dep = isl_map_apply_range(dep,
			    isl_map_reverse(
				d->to_access->extension->get_isl_map(ctx)));
	if (has_controls)
		d->controlled_relation = new pdg::IslMap(isl_map_copy(dep));
	d->relation = new pdg::IslMap(remove_all_controls(isl_map_copy(dep)));
	info->pdg->dependences.push_back(d);
	isl_map_free(dep);
	return 0;
}

/* Assuming "coa" is a (read) access, construct a map from the domain
 * of the access relation to the access relation of the corresponding
 * write.  If we are unable to determine the corresponding write, then
 * return the empty map.
 */
static __isl_give isl_map *extract_access_map(pdg::call_or_access *coa)
{
	assert(coa->type == pdg::call_or_access::t_access);
	return coa->access->extract_access_map();
}

/* Given a map from sinks to potential sources (source_map),
 * check if any parametrization is needed on the sources.
 * That is, check whether the possible filter values at the sink
 * imply that the filter values at the source are always valid.
 * If so, the source is executed whenever the sink is executed
 * and no parametrization is required.
 *
 * If there are no filters on the source, then obviously the source
 * is always executed.  Similarly, if there are no filters on the sink,
 * then there is no information to conclude anything from and so we
 * need to introduce parameters.
 *
 * The "source" domains of both source and sink are wrapped maps,
 * mapping iteration vectors to filter values.
 * We first construct a relation between the sink filter values
 * and the source filter values.  The purpose of this relation
 * is to find as much information about the source filter values
 * as possible.  We can start with the value bounds on the arrays
 * accessed in the filters as they always hold.
 * Next, we loop over the source filters and check whether there
 * is any sink filter that covers the source filter.
 * In particular, for each source filter, we construct a map
 * from the source iteration domain to a wrapped access relation,
 * representing the write access relation that corresponds to
 * the filter read access.  If we are unable to determine this
 * write access, then we continue with the next source filter.
 * Otherwise we combine the constructed map with the proto-dependence
 * (source_map) to obtain a mapping from sink iterations to
 * access relations such that there is some source iteration that
 * may be the source of the given sink iteration (based on source_map)
 * and that the filter value at this source iteration was written by
 * that access.  If the result is a subset of the mapping from the
 * sink iterations to the corresponding write access relation for some filter,
 * then we know that any constraint on this filter value also applies
 * to the source filter value.  We therefore introduce an equality
 * in our mapping from sink filter values to source filter values.
 *
 * When we apply the mapping from sink filter values to source filter values
 * to the mapping from source iterations to sink filter values, we
 * obtain a mapping from sink iterations to source filter values
 * that represents what we know about the source filter values.
 * That is, for each sink iteration in the domain of this map, if this
 * sink iteration is executed, then the actual source filter values
 * are an element of the image of the sink iteration.
 * In other words, the sink iteration is only executed if the source
 * filter values are an element of the image.
 * Mapping this relation back to the source through the proto-dependence
 * source_map, we obtain a relation from source iterations to all source
 * filter values for which any sink iteration is executed.
 * In particular, for values of the filters outside this relation,
 * no corresponding (according to source_map) sink is executed.
 * If this relation is a subset of the actual mapping from iteration
 * vectors to filter values at the source, then we know that a corresponding
 * sink is only executed when the source is executed and no parametrization
 * is required.
 */
static bool need_parametrization(__isl_keep isl_map *source_map,
	na_pair *sink_na, na_pair *source_na)
{
	bool filtered_source, filtered_sink;
	int covered;
	isl_ctx *ctx;
	isl_space *space;
	isl_map *source_value, *sink_value;
	isl_map *filter_map;
	isl_set *bounds;

	if (isl_map_plain_is_empty(source_map))
		return false;

	filtered_source = isl_set_is_wrapping(source_na->node->source->set);
	filtered_sink = isl_set_is_wrapping(sink_na->node->source->set);

	if (!filtered_source)
		return false;
	if (!filtered_sink)
		return true;

	ctx = isl_map_get_ctx(source_map);
	source_value = isl_set_unwrap(source_na->node->source->get_isl_set());
	sink_value = isl_set_unwrap(sink_na->node->source->get_isl_set());

	space = isl_space_map_from_domain_and_range(
			isl_space_range(isl_map_get_space(sink_value)),
			isl_space_range(isl_map_get_space(source_value)));
	filter_map = isl_map_universe(space);

	bounds = isl_set_universe(isl_space_set_alloc(ctx, 0, 0));
	for (int i = 0; i < source_na->node->filters.size(); ++i) {
		isl_set *bound;
		pdg::array *array;
		pdg::call_or_access *coa = source_na->node->filters[i];
		assert(coa->type == pdg::call_or_access::t_access);
		array = coa->access->array;
		if (array->value_bounds)
			bound = array->value_bounds->get_isl_set();
		else
			bound = isl_set_universe(isl_space_set_alloc(ctx, 0, 1));
		bounds = isl_set_flat_product(bounds, bound);
	}
	filter_map = isl_map_intersect_range(filter_map, bounds);

	for (int i = 0; i < source_na->node->filters.size(); ++i) {
		isl_map *map_i;
		map_i = extract_access_map(source_na->node->filters[i]);
		if (isl_map_is_empty(map_i)) {
			isl_map_free(map_i);
			continue;
		}
		map_i = isl_map_apply_range(isl_map_copy(source_map), map_i);
		for (int j = 0; j < sink_na->node->filters.size(); ++j) {
			isl_map *map_j;
			map_j = extract_access_map(sink_na->node->filters[j]);
			if (isl_map_is_subset(map_i, map_j))
				filter_map = isl_map_equate(filter_map,
					isl_dim_in, j, isl_dim_out, i);
			isl_map_free(map_j);
		}
		isl_map_free(map_i);
	}

	sink_value = isl_map_apply_range(sink_value, filter_map);
	sink_value = isl_map_apply_domain(sink_value, isl_map_copy(source_map));

	covered = isl_map_is_subset(sink_value, source_value);

	isl_map_free(source_value);
	isl_map_free(sink_value);

	return !covered;
}

extern "C" {
	static __isl_give isl_set *restrict_sources(
		__isl_keep isl_map *source_map,
		void *sink_user, void *source_user);
}

/* Given a map from sinks to potential sources (source_map),
 * check if any parametrization is needed on the sources.
 * If so, return a parametrized source space.
 * Otherwise, simply return the universe source space.
 */
static __isl_give isl_set *restrict_sources(__isl_take isl_map *source_map,
	void *sink_user, void *source_user)
{
	na_pair *sink_na = (na_pair *) sink_user;
	na_pair *source_na = (na_pair *) source_user;
	isl_set *set;
	bool need;

	set = isl_set_universe(isl_space_range(isl_map_get_space(source_map)));

	if (need_parametrization(source_map, sink_na, source_na))
		set = add_parametrization(set, source_na);

	isl_map_free(source_map);

	return set;
}

/* Does the iteration domain of any of the writes involve any filters?
 */
static bool any_filters(vector<na_pair> &writers)
{
	for (int i = 0; i < writers.size(); ++i)
		if (isl_set_is_wrapping(writers[i].node->source->set))
			return true;
	return false;
}

void find_deps(PDG* pdg, pdg::array *a, type t)
{
    isl_ctx *ctx = pdg->get_isl_ctx();
    int nparam = pdg->params.size();
    int selfinput = 0;
    int reuse = 0;
    int firstuse = 0;
    add_dep_info info = { pdg, a, t };
    isl_set *context;
    bool need_parametrization;

    switch (t) {
    case flow:
	info.dtype = pdg::dependence::flow;
	firstuse = 1;
	break;
    case anti:
	info.dtype = pdg::dependence::anti;
	break;
    case data_reuse:
	info.dtype = pdg::dependence::reuse;
	reuse = 1;
	firstuse = 1;
	break;
    case reuse_pair:
	info.dtype = pdg::dependence::reuse_pair;
	break;
    case output:
	info.dtype = pdg::dependence::output;
	break;
    }

    a->analysis_performed.push_back(new pdg::dependence_type(info.dtype));

    vector<na_pair> readers;
    vector<na_pair> writers;
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	pdg::statement *s = pdg->nodes[i]->statement;
	for (int j = 0; j < s->accesses.size(); ++j) {
	    pdg::access *access = s->accesses[j];
	    if (access->array != a)
		continue;
	    switch (t) {
	    case flow:
	    case data_reuse:
	    case anti:
		if ((access->type == pdg::access::read) ^ (t == anti))
		    readers.push_back(na_pair(node, access));
		else
		    writers.push_back(na_pair(node, access));
		break;
	    case output:
		if (access->type == pdg::access::read)
		    continue;
	    case reuse_pair:
		readers.push_back(na_pair(node, access));
		writers.push_back(na_pair(node, access));
		break;
	    }
	}
    }

    int maxsize = (selfinput || reuse) ? writers.size() + readers.size()
				       : writers.size();
    context = pdg->get_context_isl_set();
    info.precedes_level = (isl_access_level_before)
	    ((t == reuse_pair) ? precedes_level_accesses
				: precedes_level_nodes);
    need_parametrization = any_filters(writers);
    for (int i = 0; i < writers.size(); ++i) {
	writers[i].map = convert_access(&writers[i]);
    }
    for (int i = 0; i < readers.size(); ++i) {
	readers[i].map = convert_access(&readers[i]);
	readers[i].map = isl_map_intersect_domain(readers[i].map,
						isl_set_copy(context));
    }
    for (int i = 0; i < readers.size(); ++i) {
	isl_access_info *acc;
	acc = isl_access_info_alloc(isl_map_copy(readers[i].map),
				    &readers[i], info.precedes_level, maxsize);
	if (need_parametrization)
		acc = isl_access_info_set_restrict_sources(acc,
							    &restrict_sources);
	for (int j = 0; j < writers.size(); ++j)
		acc = isl_access_info_add_source(acc,
			isl_map_copy(writers[j].map), 1, &writers[j]);
	if (selfinput && writers.size()) {
	    pdg::node *readnode = readers[i].node;
	    for (int j = 0; j < readers.size(); ++j) {
		if (readers[j].node == readnode)
		    acc = isl_access_info_add_source(acc,
			isl_map_copy(readers[j].map), 1, &readers[j]);
	    }
	}
	if (reuse) {
	    for (int j = 0; j < readers.size(); ++j)
		acc = isl_access_info_add_source(acc,
		    isl_map_copy(readers[j].map), 1, &readers[j]);
	}
	isl_flow *deps = isl_access_info_compute_flow(acc);
	info.read_na_pair = &readers[i];
	isl_flow_foreach(deps, add_dep, &info);
	isl_map *no_source;
	no_source = isl_flow_get_no_source(deps, 1);
	no_source = isl_map_from_range(isl_map_domain(no_source));
	no_source = simplify_controls(no_source, &info, NULL);
	if (!isl_map_fast_is_empty(no_source) && firstuse) {
	    pdg::dependence *d = new pdg::dependence;
	    d->array = a;
	    d->type = pdg::dependence::uninitialized;
	    d->to = readers[i].node;
	    d->to_access = readers[i].access;
	    if (d->to_access->extension) {
		d->extended_relation = new pdg::IslMap(isl_map_copy(no_source));
		no_source = isl_map_apply_range(no_source,
			    isl_map_reverse(
				d->to_access->extension->get_isl_map(ctx)));
	    }
	    d->relation = new pdg::IslMap(isl_map_copy(no_source));
	    pdg->dependences.push_back(d);
	}
	isl_map_free(no_source);
	isl_flow_free(deps);
    }
    isl_set_free(context);
}

extern "C" {
	static int add_source(__isl_take isl_map *dep, int must, void *dep_user,
		void *user);
}

/* Add "dep" to a->sources, provided it is exact, and return 0.
 * Otherwise, return -1.
 */
static int add_source(__isl_take isl_map *dep, int must, void *dep_user,
	void *user)
{
	bool has_controls;
	pdg::access *a = (pdg::access *) user;

	dep = remove_redundant_controls(dep, &has_controls);
	if (has_controls) {
		isl_map_free(dep);
		return -1;
	}

	a->sources.push_back(new pdg::IslMap(dep));

	return 0;
}

extern "C" {
	static __isl_give isl_set *restrict_sources_domain_map(
		__isl_keep isl_map *source_map,
		void *sink_user, void *source_user);
}

/* Given a map from sinks to potential sources (source_map),
 * check if any parametrization is needed on the sources.
 * If so, return a parametrized source space.
 * Otherwise, simply return the universe source space.
 *
 * This function is similar to restrict_sources.
 * The main difference is that it is used from within find_sources,
 * which encodes the entire access relation into the domains of
 * the access relations passed to isl_access_info_compute_flow.
 * That is, the access relations passed to isl_access_info_compute_flow
 * are the result of applying isl_map_range_map to the original access
 * relations.  We have to undo this transformation before calling
 * need_parametrization and add_parametrization and reapply it
 * to the result of add_parametrization.
 */
static __isl_give isl_set *restrict_sources_domain_map(
	__isl_keep isl_map *source_map, void *sink_user, void *source_user)
{
	na_pair *sink_na = (na_pair *) sink_user;
	na_pair *source_na = (na_pair *) source_user;
	isl_space *space;
	isl_map *source_domain_map, *sink_domain_map;
	isl_set *set;

	space = isl_space_range(isl_map_get_space(source_map));
	space = isl_space_unwrap(space);
	source_domain_map = isl_map_domain_map(isl_map_universe(space));
	space = isl_space_domain(isl_map_get_space(source_map));
	space = isl_space_unwrap(space);
	sink_domain_map = isl_map_domain_map(isl_map_universe(space));

	set = isl_set_universe(isl_space_range(isl_map_get_space(source_map)));
	set = isl_set_apply(set, isl_map_copy(source_domain_map));

	source_map = isl_map_apply_range(source_map,
				isl_map_copy(source_domain_map));
	source_map = isl_map_apply_domain(source_map, sink_domain_map);

	if (need_parametrization(source_map, sink_na, source_na))
		set = add_parametrization(set, source_na);

	set = isl_set_apply(set, isl_map_reverse(source_domain_map));

	isl_map_free(source_map);

	return set;
}

/* Find the sources of (read) access "a" in node "node".
 * If they are complete (no uninitialized accesses) and exact,
 * then put them in a->sources.  Otherwise, discard them.
 *
 * We look for all writes that write to the same array,
 * perform dependence analysis and then check whether
 * the result is complete and exact.
 *
 * The sources record not only the node iteration, but also
 * the index of the array element.  We therefore apply
 * isl_map_range_map to the access relations, to obtain
 * a relation from the access (iteration -> element)
 * to the array element and feed that to the dependence analysis engine.
 */
void find_sources(pdg::PDG* pdg, pdg::node *node, pdg::access *a)
{
	isl_set *context;
	isl_map *no_source;
	isl_access_info *acc;
	isl_flow *deps;
	vector<na_pair> writers;
	na_pair reader(node, a);
	add_dep_info info = { pdg, a->array, flow, pdg::dependence::flow };

	info.precedes_level = (isl_access_level_before) precedes_level_nodes;

	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = pdg->nodes[i]->statement;
		for (int j = 0; j < s->accesses.size(); ++j) {
			pdg::access *access = s->accesses[j];
			if (access->array != a->array)
				continue;
			if (access->type != pdg::access::write)
				continue;
			writers.push_back(na_pair(node, access));
		}
	}

	context = pdg->get_context_isl_set();
	for (int i = 0; i < writers.size(); ++i)
		writers[i].map = isl_map_range_map(convert_access(&writers[i]));
	reader.map = isl_map_range_map(convert_access(&reader));
	reader.map = isl_map_intersect_params(reader.map, context);

	acc = isl_access_info_alloc(isl_map_copy(reader.map),
				&reader, info.precedes_level, writers.size());
	for (int j = 0; j < writers.size(); ++j)
		acc = isl_access_info_add_source(acc,
			isl_map_copy(writers[j].map), 1, &writers[j]);

	if (any_filters(writers))
		acc = isl_access_info_set_restrict_sources(acc,
						&restrict_sources_domain_map);

	deps = isl_access_info_compute_flow(acc);
	no_source = isl_flow_get_no_source(deps, 1);
	if (isl_map_plain_is_empty(no_source)) {
		if (isl_flow_foreach(deps, add_source, a) < 0) {
			for (int i = 0; i < a->sources.size(); ++i)
				delete a->sources[i];
			a->sources.resize(0);
		}
	}
	isl_map_free(no_source);
	isl_flow_free(deps);
}

/* Compute the sources (if possible) for all the filters in all the
 * nodes in "pdg".
 */
void compute_filter_sources(pdg::PDG *pdg)
{
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		int n_filter = node->filters.size();

		for (int j = 0; j < n_filter; ++j) {
			pdg::call_or_access *coa;
			pdg::access *access;

			coa = node->filters[j];
			assert(coa->type == pdg::call_or_access::t_access);
			access = coa->access;

			find_sources(pdg, node, access);
		}
	}
}

static int precedes_level_nodes(na_pair *first, na_pair *second)
{
    int d = 0;
    int cmp = 0;
    for (int i = 0; i < first->node->prefix.size(); ++i) {
	if (i >= second->node->prefix.size())
	    break;
	cmp = first->node->prefix[i] - second->node->prefix[i];
	if (cmp)
	    break;
	if (first->node->prefix[i] == -1)
	    ++d;
    }
    return 2*d + (cmp<0);
}

static int precedes_level_accesses(na_pair *first, na_pair *second)
{
    int cmp = 0;
    int d = 0;
    for (int i = 0; i < first->node->prefix.size(); ++i) {
	if (i >= second->node->prefix.size())
	    break;
	cmp = first->node->prefix[i] - second->node->prefix[i];
	if (cmp)
	    break;
	if (first->node->prefix[i] == -1)
	    ++d;
    }
    /* same node; now compare accesses */
    if (!cmp)
	cmp = first->access->nr - second->access->nr;
    return 2*d + (cmp<0);
}

}
